use structopt::StructOpt;
use tokio::net::TcpListener;

use quic_demo::prelude::*;

#[derive (Debug, StructOpt)]
struct Opt {
	#[structopt (long)]
	relay_addr: Option <String>,
	#[structopt (long)]
	local_tcp_port: Option <u16>,
	#[structopt (long)]
	client_id: Option <u8>,
	#[structopt (long)]
	server_id: Option <u8>,
}

#[tokio::main]
async fn main () -> anyhow::Result <()> {
	tracing_subscriber::fmt::init ();
	
	let opt = Opt::from_args ();
	let local_tcp_port = opt.local_tcp_port.unwrap_or (30381);
	
	let server_cert = tokio::fs::read ("quic_server.crt").await?;
	let relay_addr = opt.relay_addr.unwrap_or_else (|| String::from ("127.0.0.1:30380")).parse ()?;
	let endpoint = make_client_endpoint ("0.0.0.0:0".parse ()?, &[&server_cert])?;
	
	trace! ("Connecting to relay server");
	
	let client_id = opt.client_id.unwrap_or (42);
	
	let quinn::NewConnection {
		connection,
		..
	} = protocol::p2_connect_to_p3 (&endpoint, &relay_addr, client_id).await?;
	
	let server_id = opt.server_id.unwrap_or (43);
	
	let listener = TcpListener::bind (("127.0.0.1", local_tcp_port)).await?;
	
	trace! ("Accepting local TCP connections from P1");
	
	loop {
		let (tcp_socket, _) = listener.accept ().await?;
		let connection = connection.clone ();
		
		tokio::spawn (async move {
			let (local_recv, local_send) = tcp_socket.into_split ();
			
			debug! ("Starting PTTH connection");
			
			let (relay_send, relay_recv) = protocol::p2_connect_to_p5 (&connection, server_id).await?;
			
			trace! ("Relaying bytes...");
			
			let ptth_conn = quic_demo::connection::NewConnection {
				local_send,
				local_recv,
				relay_send,
				relay_recv,
			}.build ();
			
			ptth_conn.wait_for_close ().await?;
			
			debug! ("Ended PTTH connection");
			
			Ok::<_, anyhow::Error> (())
		});
	}
}
